#include <ScatterWeb.System.h>
#include <ScatterWeb.CC1020.h>
#include <string.h>
#include "ScatterWeb.Process.h"

// naechste position
uint16_t next_pos = 0;
// das DEBUG-flag
uint16_t DEBUG = 0;
// pakete angekommen
uint16_t received = 0;
// uid, route request packet
uint16_t seq_no = 1;
// versuchscache
uint16_t trycache[20];
// cached uid's
uint16_t cacheduids[20];
// targets for cached data
uint16_t targetcache[20];
// datencache
uint16_t datacached[10];
// reached packets
uint16_t ids[10];
// angeforderte pakete
uint16_t requested = 0;
// angekommene antworten
uint16_t replies[20];


bool Process_radioHandler(struct netpacket_handler_args* args) {
	/*
	 Notes: * The received packet data becomes invalid when this handler is completed.
	 */

	// Pass packet through command interface
	if (Cmd_radioHandler(args)) {
		// packet handled by messaging module
	}

	netpacket_send_args_t npsargs;
	Net_sendArgsInit(&npsargs);
	// intercept packets for other nodes here
	// ueberpruefe die Node-Ids. Meine sind 61-70.
	if (args->cpaResult == ADRCOMP_OTHER && args->netheader->from >= 61
			&& args->netheader->from <= 70 && args->netheader->to >= 61
			&& args->netheader->to <= 70) {

		// ab hier wissen wir, dass nur meine Node-Ids behandelt werden.
		// das zu untersuchende Paket ist eine Anfrage
		if (args->netheader->type == REQUEST) {
			// Flag fuer zuvor erhaltenes Paket
			uint16_t ZUVOR = 0;
			// Paket
			uint16_t result;
			// flache Paketkopie
			memcpy(&route, args->payload, args->payload_size);
			// dursuche Array nach bereits vorhandener ID
			uint16_t p;
			for (p = 0; p < 1 + seq_no; p++) {
				if (ids[p] == route.id) {
					ZUVOR = 1;
					result = ids[p];
				}
			}
			// Verwerfe Anfrage?
			if (ZUVOR) {
				if (DEBUG){
					printf("Verwerfe Anfrage von %u.\r\n",route.source);
				}
				return false;
			}
			// Verwerfe Antwort, oder antworte falls das Ziel erreicht wurde
			if (npsargs.netheader.from == route.target) {
				p = 0;
				uint16_t ZUVOR = 0;
				for (p = 0; p < 1 + next_pos; p++) {
					if (route.id == replies[p]) {
						ZUVOR = 1;
					}
				}
				// sende bereits erhaltenes Paket zur�ck
				if (ZUVOR == 0) {
					replies[next_pos = (1 + next_pos) % 20] = route.id;
					send_reply(args);
				}
				// verwerfe sonst
				else {
					if (DEBUG){
						printf("Verwerfe Antwort von %u.\r\n",route.target);
					}
					return false;
				}
			}
			// sende erneut falls Ziel nicht erreicht wurde
			else {
				send_request(route.target, 1);
			}
		}
	}
	else {
		// only receive messages from nodes out of this range
		if (args->netheader->from <= 70 && args->netheader->to <= 70
				&& args->netheader->from >= 61 && args->netheader->to >= 61) {
			// process packets for this node here
			switch (args->netheader->type) {

			case PING_REQUEST_PACKET:
				Net_sendPong(args->netheader->from);
				break;

			case PING_REPLY_PACKET:
				LOG_LOW("PONG from %u", args->netheader->from);
				break;

			case DATA:
				// Kopie
				memcpy(&data, args->payload, args->payload_size);
				send_data(args, 0);
				break;

			case REPLY:
				// Kopie
				memcpy(&route, args->payload, args->payload_size);
				send_reply(args);
				break;

			default:
				return false;
			}
		}

		return true;
	}
	return false;
}
// Hilfsfunktion zum clearen des Cacheposition
void clearcachepos(uint16_t p){
	trycache[p]=0;
}
/**
 * Erstellen und Versenden des Antwort-Pakets
 */
void send_reply(struct netpacket_handler_args* args) {
	netpacket_send_args_t npsargs;
	Net_sendArgsInit(&npsargs);

	if (route.target == npsargs.netheader.from) {
		if (DEBUG){
			printf("Anfrage von %u hat ihr Ziel %u erreicht. \r\n",route.source, route.target);
		}
		route.header[route.header[0]] = route.target;
		route.header[route.header[0] + 1] = 0;
	}
	if (1 == route.header[0]) {
		if (DEBUG){
			printf("Ziel %u hat seinen Start %u erreicht.\r\n",route.target, route.source);
		}
		// sende Daten
		send_data(args, 1);

		uint16_t p;
		for (p = 0; p < 20; p++) {
			if (trycache[p] < 4 && 0 != cacheduids[p]) {
				// setze den payload-buffer
				data.data = datacached[p];
				npsargs.payload[0].buffer = &data;
				npsargs.payload[0].size = sizeof(data);
				// setze den netheader
				npsargs.netheader.type = DATA;
				npsargs.netheader.flags = 0;
				npsargs.netheader.to = targetcache[p];
				Net_send(&npsargs);
			} else {
				cacheduids[p] = 0;
			}
			trycache[p]++;
		}
		return true;
	}
	else {
		if (DEBUG){
			printf("Durchgereicht von %u nach Ziel %u",npsargs.netheader.from, route.target);
		}
		route.header[0]--;
	}
	// setze den payload-buffer
	npsargs.payload[0].buffer = &route;
	npsargs.payload[0].size = sizeof(route);
	// setze den netheader
	npsargs.netheader.type = REPLY;
	npsargs.netheader.to = route.header[route.header[0]];
	npsargs.netheader.flags = 0;
	Net_send(&npsargs);
}
/**
 * Sende Anfrage an Ziel mit flag (neues Paket, oder geaendertes)
 */
void send_request(uint16_t target, uint16_t flag) {
	netpacket_send_args_t npsargs;
	Net_sendArgsInit(&npsargs);

	// setze netheader
	npsargs.netheader.type = REQUEST;
	npsargs.netheader.to = ADRCOMP_BROADCAST;
	npsargs.netheader.flags = 0;
	// erstelle Paket und versende
	if (!flag) {
		// Zaehlvariable
		uint16_t p;
		// cache Variable
		uint16_t FREI = 0;
		for (p = 0; 0 == FREI && p < 20; p++) {
			// freie Speicherstelle gefunden
			if (0 == cacheduids[p]) {
				FREI = 1;
				cacheduids[p] = seq_no;datacached[p] = seq_no;targetcache[p] = target;
				clearcachepos(p);
			}
		}
		// Abbruch, da der Cache voll ist
		if (!FREI){
			printf("Cache ist voll Abbruch!");
			return false;
		}

		next_pos = next_pos % 20;
		ids[next_pos] = seq_no;
		next_pos++;
		// setze Routeninfos vom netheader
		route.id = seq_no++;
		route.source = npsargs.netheader.from;
		route.target = target;
		route.header[1] = npsargs.netheader.from;
		route.header[0] = 2;
		if (DEBUG)
			printf("Sende Paketnummer %u von %u.\r\n",route.target, route.source);
	}
	// Anedere das Paket
	else {
		next_pos = next_pos % 20;
		ids[next_pos] = route.id;
		next_pos++;
		// setze den header um
		route.header[route.header[0]] = npsargs.netheader.from;++route.header[0];
	}
	// setze wieder den payload-buffer
	npsargs.payload[0].buffer = &route;
	npsargs.payload[0].size = sizeof(route);
	// sende Paket, neues oder geaendertes, je nach dem
	Net_send(&npsargs);
}
/**
 * Sende Daten mit flag (neu, geaendert)
 */
void send_data(struct netpacket_handler_args* args, uint16_t flag) {
	netpacket_send_args_t npsargs;
	Net_sendArgsInit(&npsargs);
	// neues Paket
	if (flag) {
		uint16_t p = 0;
		// suche die richtige Stelle
		while (cacheduids[p] != route.id) {
			p++;
		}
		// Position p gefunden
		// entferne aus dem cache
		cacheduids[p] = 0;
		// setze die Daten
		data.data = datacached[p];
		uint16_t i;
		for (i = 0; i < 8; i++) {
			data.header[i] = route.header[i];
		}
		++data.header[0];
	}
	// geandertes Paket
	else {
		++data.header[0];
		if (data.header[data.header[0]] == 0) {
			received++;
			return true;
		}
	}
	// setze payload-buffer
	npsargs.payload[0].buffer = &data;
	npsargs.payload[0].size = sizeof(data);
	// setze netheader
	npsargs.netheader.type = DATA;
	npsargs.netheader.to = data.header[data.header[0]];
	npsargs.netheader.flags = 0;
	// sende Paket, neues oder geaendertes, je nach dem
	Net_send(&npsargs);
}
void reroute() {
	uint16_t p;
	// durchsuche das array
	for (p = 0; p < 20; p++) {
		if (trycache[p] < 4 && 0 != cacheduids[p]) {
			// die uid wurde gefunden, sende die Anfrage
			trycache[p]++;
			send_request(targetcache[p], 1);
			if (DEBUG){
				printf("Packet wird nochmal gesendet %u.\r\n", cacheduids[p]);
			}
		} else
			cacheduids[p] = 0;
	}
	Timers_add(3 * 1024, reroute, 0xffff);
}
/**
 * @brief unused sensorHandler function
 */
void Process_sensorHandler(void* sdata) {
}
/**
 * Predefinde Init function
 */
void Process_init() {
	System_registerCallback(C_RADIO, (fp_vp) Process_radioHandler);
	System_registerCallback(C_SENSOR, (fp_vp) Process_sensorHandler);
	reroute();
}
/**
 * DEBUG
 */
COMMAND( DEBUG, CMDFLAG_SERIAL, cmdargs) {
	DEBUG = String_parseUint16(cmdargs->args, NULL);
}

/**
 * Send Packet
 * param:	Target
 */
COMMAND( SENDPACKET, CMDFLAG_SERIAL, cmdargs) {
	send_request(String_parseUint16(cmdargs->args, NULL), 0);
}

/**
 * Ping a node
 * param:	Target
 */
COMMAND( ping, 0, cmdargs )
{
	uint16_t node;

	node = String_parseUint16( cmdargs->args, NULL );
	if ( node> 0 ) {
		Net_sendPing((netaddr_t)node);
	}
}

COMMAND( echoargs, CMDFLAG_SERIAL, cmdargs) {
	uint16_t i;
	uint16_t count = Cmd_SplitArgs(cmdargs->args);
	char* arg = (char*) cmdargs->args;

	for (i = 1; i <= count; i++) {
		printf("%u: %*s\n", i, strlen(arg), arg);
		arg = (char*) Cmd_GetNextArg(arg);
	}
	snprintf(cmdargs->response_args, cmdargs->response_bufsize,
			"%u args found", count);
}

// Software header.
APP(
		APPH_STATIC_APP,
		__APP_NAME,
		1, 0
);
